#!/usr/bin/python
import os
from pdb import pm
from miasm.analysis.sandbox import Sandbox_Win_x86_32
from miasm.core.locationdb import LocationDB
from miasm.jitter.csts import *
from miasm.os_dep.common import set_win_str_w
from miasm.os_dep.win_api_x86_32 import FLS, winobjs

fls = FLS()
kernel32_FlsAlloc = fls.kernel32_FlsAlloc
kernel32_FlsSetValue = fls.kernel32_FlsSetValue
kernel32_FlsGetValue = fls.kernel32_FlsGetValue

GetEnvironmentStringsW_addr = None
all_heap = []
ErrorMode = 1

def kernel32_GetEnvironmentStringsW(jitter):
    """
    LPWCH GetEnvironmentStringsW()
    """
    global GetEnvironmentStringsW_addr
    ret_ad, args = jitter.func_args_stdcall([])

    # If GetEnvironmentStrings isn't allocated yet
    if not GetEnvironmentStringsW_addr:
        alloc_addr = winobjs.heap.next_addr(50)
        winobjs.allocated_pages[alloc_addr] = (alloc_addr, 50)
        jitter.vm.add_memory_page(alloc_addr, PAGE_READ | PAGE_WRITE, b"\x00"*50)
        set_win_str_w(jitter, alloc_addr, "Hello")
        GetEnvironmentStringsW_addr = alloc_addr
    jitter.func_ret_stdcall(ret_ad, GetEnvironmentStringsW_addr)

def kernel32_FreeEnvironmentStringsW(jitter):
    ret_ad, args = jitter.func_args_stdcall(["lpszEnvironmentBlock"])
    jitter.func_ret_stdcall(ret_ad, 1)
    jitter.running = True
    return True

def kernel32_LCMapStringEx(jitter):
    """
    int LCMapStringEx(
        LPCWSTR lpLocaleName,
        [LocaleMappingFlags] dwMapFlags,
        LPCWSTR lpSrcStr,
        int cchSrc,
        LPWSTR lpDestStr,
        int cchDest,
        LPNLSVERSIONINFO lpVersionInformation,
        LPVOID lpReserved,
        LPARAM lParam
    )
    """
    ret_ad, args = jitter.func_args_stdcall(["lpLocaleName", "dwMapFlags", "lpSrcStr", "cchSrc", "lpDestStr", "cchDest", "lpVersionInformation", "lpReserved", "lParam"])
    jitter.func_ret_stdcall(ret_ad, 0x100)

def kernel32_SetUnhandledExceptionFilter(jitter):
    """
    LPTOP_LEVEL_EXCEPTION_FILTER SetUnhandledExceptionFilter(
        LPTOP_LEVEL_EXCEPTION_FILTER lpTopLevelExceptionFilter
    )
    """
    ret_ad, args = jitter.func_args_stdcall(["lpTopLevelExceptionFilter"])
    jitter.func_ret_stdcall(ret_ad, 0x0)

def kernel32_HeapAlloc(jitter):
    """
    LPVOID HeapAlloc(
        HANDLE hHeap,
        [HEAP_FLAGS] dwFlags,
        SIZE_T dwBytes
    )
    """
    ret_ad, args = jitter.func_args_stdcall(["heap", "flags", "size"])
    alloc_addr = winobjs.heap.alloc(jitter, args.size, cmt=hex(ret_ad))
    all_heap.append([alloc_addr, args.size])
    jitter.func_ret_stdcall(ret_ad, alloc_addr)

def kernel32_HeapSize(jitter):
    """
    SIZE_T HeapSize(
        HANDLE hHeap,
        [HEAP_FLAGS] dwFlags,
        LPCVOID lpMem
    )
    """
    ret_ad, args = jitter.func_args_stdcall(["hHeap", "dwFlags", "lpMem"])
    ret_val = 0
    for heap in all_heap:
        if heap[0] == args.lpMem:
            ret_val = heap[1]
    jitter.func_ret_stdcall(ret_ad, ret_val)

def gdi32_SelectObject(jitter):
    """
    HGDIOBJ SelectObject(
        HDC hdc,
        HGDIOBJ hgdiobj
    )
    """
    ret_ad, _args = jitter.func_args_stdcall(["hdc", "hgdiobj"])
    jitter.cpu.EDI = 1
    ret_value = 0
    jitter.func_ret_stdcall(ret_ad, ret_value)
    jitter.running = True
    return True

def kernel32_Module32First(jitter):
    """
    BOOL Module32First(
        HANDLE hSnapshot,
        LPMODULEENTRY32 lpme
    )
    """
    ret_ad, _args = jitter.func_args_stdcall(["hSnapshot", "lpme"])
    jitter.func_ret_stdcall(ret_ad, 1)
    jitter.running = True
    return True

def kernel32_SetErrorMode(jitter):
    """
    [SEM_FLAGS] SetErrorMode(
        [SEM_FLAGS] uMode
    )
    """
    global ErrorMode
    ret_ad, args = jitter.func_args_stdcall(["uMode"])
    ret_value = ErrorMode
    ErrorMode = args.uMode
    jitter.func_ret_stdcall(ret_ad, ret_value)

def kernel32_VirtualFree(jitter):
    ret_ad, args = jitter.func_args_stdcall(["lpvoid", "dwsize", "alloc_type"])
    all_mem = jitter.vm.get_all_memory()
    for region in all_mem:
        if region != args.lpvoid:
            continue
        region_data = all_mem[region]["data"]
        if region_data[:2] == b"MZ":
            print(f"PE unpacked : '{out_fname}' (Region : 0x{region:x})")
            f = open(out_fname, "wb")
            f.write(region_data)
            f.close()
            exit()

#
#   Stage 3 (b1a73a20fb4a434c77aeabd506a0995b8c7c22bbbb4824584cdecb63ca198291)
#

def kernel32_GetTickCount(jitter):
    ret_ad, _ = jitter.func_args_stdcall(0)
    winobjs.tickcount += 1
    # AntiEmulation @0x00401cae (b1a73a20fb4a434c77aeabd506a0995b8c7c22bbbb4824584cdecb63ca198291)
    if ret_ad == 0x00401cb4:
        ret_ad = 0x00401d16
    jitter.func_ret_stdcall(ret_ad, winobjs.tickcount)

def kernel32_GetLastError(jitter):
    ret_ad, _ = jitter.func_args_stdcall(0)
    # AntiEmulation @0x00401708 (b1a73a20fb4a434c77aeabd506a0995b8c7c22bbbb4824584cdecb63ca198291)
    if ret_ad == 0x40170e: # anti emu
        jitter.cpu.EDI = 0x1
    jitter.func_ret_stdcall(ret_ad, winobjs.lastwin32error)

def kernel32_SetLastError(jitter):
    ret_ad, args = jitter.func_args_stdcall(["errcode"])
    # AntiEmulation @0x004017ca (b1a73a20fb4a434c77aeabd506a0995b8c7c22bbbb4824584cdecb63ca198291)
    if ret_ad == 0x4017d0: # anti emu
        if jitter.cpu.EDI > 0x62:
            ret_ad = 0x004017e3
        else:
            jitter.cpu.EDI = 0x62
    winobjs.lastwin32error = args.errcode
    jitter.func_ret_stdcall(ret_ad, 0)

#
# ANTI-EMU
#

def jmp_0x00401bda(jitter):
    jitter.pc = 0x00401bda
    jitter.running = True
    return True

def set_esi_0x674db(jitter):
    jitter.cpu.ESI = 0x674db
    jitter.running = True
    return True

# Parse arguments
parser = Sandbox_Win_x86_32.parser(description="PE sandboxer")
parser.add_argument("filename", help="PE Filename")
options = parser.parse_args()
options.usesegm = True
options.use_windows_structs = True
options.load_hdr = True
options.dependencies = True

# Unpacked filename output
bname, fname = os.path.split(options.filename)
fname = os.path.join(bname, fname.replace('.', '_'))
out_fname = fname + '_miasm_unpacked.bin'

# Create sandbox
loc_db = LocationDB()
sb = Sandbox_Win_x86_32(loc_db, options.filename, options, globals())

# Breakpoints
sb.jitter.add_breakpoint(0x401b68, jmp_0x00401bda)  # anti-emu GetCurrentProcess & GetCurrentProcessId
sb.jitter.add_breakpoint(0x401829, set_esi_0x674db) # anti-emu GetLastError
sb.jitter.add_breakpoint(0x40184a, set_esi_0x674db) # anti-emu GetLastError

# Memory allocation
sb.jitter.vm.add_memory_page(0x7FFFF000, PAGE_READ | PAGE_WRITE, b"\xff"*4) # SEH

# Run
sb.run()

assert(sb.jitter.running is False)